"""
Testing the "run" functions in the pipeline elements.
"""
import pytest
pytest.mark.skip(reason='All tests still WIP')

# Todo: data needed: an experiment object
# Todo: data needed: an sqlite database
# Todo: data needed: a Shelf storage
class TestPipeline:
    def test_experiment(self):
        pass

    def test_omero_experiment(self):
        pass

    def test_tiler(self):
        pass

    def test_baby_client(self):
        pass

    def test_baby_runner(self):
        pass

    def test_pipeline(self):
        pass

